import numpy as np

MAX_STATE = 9


def generate_emission_matrix():
    # Emission matrix. E[x, y] = Pr(y | x)
    E = np.zeros((MAX_STATE + 1, MAX_STATE + 1))
    for x in range(MAX_STATE + 1):
        E[x, max(x - 1, 0)] += 1 / 3
        E[x, x] += 1 / 3
        E[x, min(x + 1, MAX_STATE)] += 1 / 3
    assert np.allclose(np.sum(E, axis=1), np.ones(MAX_STATE + 1))
    return E


def generate_transition_matrix():
    # Transition matrix. P[x,y] = Pr( y | x)
    P = np.zeros((MAX_STATE + 1, MAX_STATE + 1))
    for x in range(MAX_STATE + 1):
        P[x, max(x-1, 0)] += 1 / 4
        P[x, x] += 1 / 2
        P[x, min(x + 1, MAX_STATE)] += 1 / 4
    assert np.allclose(np.sum(P, axis=1), np.ones(MAX_STATE + 1))
    return P


def filter(P, E, ys):
    pi = np.ones(MAX_STATE + 1) / (MAX_STATE + 1)
    for yn in ys:
        # Compute const * P(yn | xn) * Sum_{x_{n-1}} P(x_{n-1} | h) P(x_n | x_{n-1})
        pi = np.multiply(E[:, yn], pi @ P)
        # Normalise the probability vector.
        pi = pi / np.sum(pi)
    return pi

E = generate_emission_matrix()
P = generate_transition_matrix()

print(filter(P, E, [3, 3, 4, 5, 4, 3, 3, 2, 2, 1, 2, 1, 2, 2]))
# Expected output: [0. 0.3749904  0.41518442 0.20982518 0. 0. 0. 0. 0. 0.]

print(filter(P, E, [3, 3, 4, 9]))
# Expected output: [nan nan nan nan nan nan nan nan nan nan]
